/***************************************************************************
                          kplayernodeview.h
                          -----------------
    begin                : Mon Apr 18 2005
    copyright            : (C) 2005-2008 by kiriuja
    email                : http://kplayer.sourceforge.net/email.html
 ***************************************************************************/

/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation, either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifndef KPLAYERNODEVIEW_H
#define KPLAYERNODEVIEW_H

#include <kdialog.h>
#include <qdockwidget.h>
#include <qsplitter.h>
#include <qtreeview.h>
#include <qvalidator.h>

#include "kplayeractionlist.h"
#include "kplayernode.h"
#include "ui_kplayerpropertiesdevice.h"

class KPlayerDeviceDialog;
class KPlayerContainerActionList;
class KPlayerNodeView;
class KPlayerListView;
class KPlayerTreeView;
class KPlayerLibrary;
class KPlayerLibraryWindow;
class KPlayerPlaylist;

/**Node name validator.
  *@author kiriuja
  */
class KPlayerNodeNameValidator : public QValidator
{
public:
  /** Constructor. References and stores the given container node. */
  KPlayerNodeNameValidator (KPlayerContainerNode* node, QObject* parent = 0);
  /** Destructor. Releases the container node. */
  virtual ~KPlayerNodeNameValidator();

  /** Validates node name. */
  virtual QValidator::State validate (QString& input, int& position) const;

protected:
  /** Container node. */
  KPlayerContainerNode* m_node;
};

/** The KPlayer device creation dialog page.
  *@author kiriuja
  */
class KPlayerPropertiesDevice : public QFrame, public Ui_KPlayerPropertiesDevicePage
{
  Q_OBJECT

public:
  /** Constructor. */
  KPlayerPropertiesDevice (QWidget* parent = 0);
  /** Destructor. */
  virtual ~KPlayerPropertiesDevice();

  /** Device dialog. */
  KPlayerDeviceDialog* parent (void)
    { return (KPlayerDeviceDialog*) QFrame::parent(); }

  /** Adds a new device node. */
  void addDevice (void);

protected slots:
  /** Enables or disables the OK button. */
  void pathChanged (const QString&);
  /** Shows or hides channel list and file options. */
  void typeChanged (int);

protected:
  /** Devices node. */
  KPlayerContainerNode* m_node;
};

/** The KPlayer device creation dialog.
  *@author kiriuja
  */
class KPlayerDeviceDialog : public KDialog
{
  Q_OBJECT

public:
  /** Constructor. */
  KPlayerDeviceDialog (QWidget* parent = 0);
  /** Destructor. */
  virtual ~KPlayerDeviceDialog();

  /** Adds a new device node. */
  void addDevice (void)
    { ((KPlayerPropertiesDevice*) mainWidget()) -> addDevice(); }
};

/** History entry.
  * @author kiriuja
  */
class KPlayerHistoryEntry
{
public:
  /** Entry URL. */
  KUrl m_url;
  /** Entry name. */
  QString m_name;
  /** List of expanded items. */
  KUrl::List m_expanded;
  /** List of selected item containers. */
  KUrl::List m_selected;
  /** List of selected item IDs. */
  QStringList m_selected_ids;
  /** Current item container. */
  KUrl m_current;
  /** Current item ID. */
  QString m_current_id;
  /** Scrolling position. */
  QPoint m_position;
};

/** History.
  * @author kiriuja
  */
class KPlayerHistory : public QObject
{
  Q_OBJECT

public:
  KPlayerHistory (QObject* parent = 0);

  const QList<KPlayerHistoryEntry>& entries (void) const
    { return m_entries; }
  int currentIndex (void) const
    { return m_current; }

  bool canGoBack (void) const
    { return m_current > 0; }
  bool canGoForward (void) const
    { return m_current < m_entries.count() - 1; }

  void goBack (void);
  void goForward (void);
  void goTo (int index);

  /** Applies current history entry, expanding and selecting items and scrolling the view. */
  void apply (KPlayerListView* view);
  /** Stores expanded and selected items and scroll position to the current history entry. */
  void store (KPlayerListView* view);

signals:
  /** Emitted when the node is selected for navigation. */
  void navigate (KPlayerContainerNode* node);
  /** Emitted when the history is changed. */
  void updated (void);

protected:
  QList<KPlayerHistoryEntry> m_entries;
  /** Current history entry index. */
  int m_current;
  /** Navigation indicator. */
  bool m_navigating;
};

/**History action list.
  *@author kiriuja
  */
class KPlayerHistoryActionList : public KPlayerActionList
{
  Q_OBJECT

public:
  /** Constructor. */
  KPlayerHistoryActionList (const KPlayerHistory* history, const KLocalizedString& text, const KLocalizedString& status,
    const KLocalizedString& whatsthis, QObject* parent, const QString& name);
  /** Destructor. */
  virtual ~KPlayerHistoryActionList();

public slots:
  /** Updates the action list. */
  void update (void);

signals:
  /** Emitted when the URL is selected. */
  void activated (int index);

protected:
  /** Selects the item with the given index by emitting the activated signal. */
  virtual void actionActivated (QAction* action, int index);

  /** History. */
  const KPlayerHistory* m_history;
  /** Index of the bottom entry. */
  int m_bottom;
};

/**Node attributes.
  *@author kiriuja
  */
class KPlayerNodeAttributes : public QObject
{
  Q_OBJECT

public:
  /** Sets up the node attributes. */
  KPlayerNodeAttributes (QObject* parent);
  /** Cleans up the node attributes. */
  virtual ~KPlayerNodeAttributes();

  /** Initializes the attributes. */
  void initialize (bool directory);

  /** Configuration. */
  KConfig* config (void) const
    { return KPlayerEngine::engine() -> config(); }

  /** Returns the number of attributes. */
  int attributeCount (void) const
    { return attributeNames().count(); }
  /** Returns the list of attribute names. */
  const QStringList& attributeNames (void) const
    { return m_attribute_names; }

  /** Returns the list of available attribute names. */
  const QStringList& availableAttributes (void) const
    { return m_available_attributes; }

  /** Returns the list of editable attribute names. */
  const QStringList& editableAttributes (void) const
    { return m_editable_attributes; }

  /** Returns the list of all possible attributes. */
  const QStringList& attributeOrder (void) const
    { return m_attribute_order; }

  /** Returns the attribute states. */
  const QMap<QString, bool>& attributeStates (void) const
    { return m_attribute_states; }

  /** Returns the attribute counts. */
  const KPlayerPropertyCounts& attributeCounts (void) const
    { return m_attribute_counts; }

  /** Toggles the column state. */
  void toggleState (const QString& name);

  /** Returns the column index of the editable attribute. Adds the column if needed. */
  int editableIndex (int index);

  /** Moves attribute to another position. */
  void move (int from, int to);

signals:
  /** Emitted when the attribute is inserted. */
  void inserted (const QString& name);
  /** Emitted when the attribute is removed. */
  void removed (const QString& name);
  /** Emitted when the attribute list is changed. */
  void updated (void);

public slots:
  /** Updates the attribute counts, add and removes columns as necessary, keeping column widths. */
  void update (const KPlayerPropertyCounts& added, const KPlayerPropertyCounts& removed);

protected:
  /** Returns whether the column with the given name has to be shown. */
  bool showColumn (const QString& name) const
    { return m_column_states.contains (name) && m_column_states [name]; }
  /** Inserts an attribute into the list according to the order. */
  void insert (QStringList& list, const QString& name);

  /** Names of visible node attributes. */
  QStringList m_attribute_names;
  /** Counts of visible node attributes. */
  KPlayerPropertyCounts m_attribute_counts;
  /** Names of available node attributes. */
  QStringList m_available_attributes;
  /** Names of editable node attributes. */
  QStringList m_editable_attributes;
  /** Names of all possible node attributes. */
  QStringList m_attribute_order;
  /** Attribute states. */
  QMap<QString, bool> m_attribute_states;
  /** Column states. */
  QMap<QString, bool> m_column_states;
  /** Directory indicator. */
  bool m_directory;
};

/**Node model.
  *@author kiriuja
  */
class KPlayerNodeModel : public QAbstractItemModel
{
  Q_OBJECT

public:
  /** Sets up the node model. */
  KPlayerNodeModel (QObject* parent = 0);
  /** Cleans up the node model. */
  virtual ~KPlayerNodeModel();

  /** Initializes the model. */
  virtual void initialize (void);
  /** Frees up resources. */
  virtual void terminate (void);

  /** Returns the parent view widget. */
  KPlayerNodeView* parent (void) const
    { return (KPlayerNodeView*) QObject::parent(); }

  /** Configuration. */
  KConfig* config (void) const
    { return KPlayerEngine::engine() -> config(); }

  /** Returns whether container items are expandable. */
  bool expandable (void) const
    { return m_expandable; }

  KPlayerNode* nodeFromIndex (const QModelIndex& index) const
    { return index.isValid() ? (KPlayerNode*) index.internalPointer() : KPlayerNode::root(); }
  QModelIndex indexForNode (KPlayerNode* node) const;

  KPlayerNodeList nodesFromIndexList (const QModelIndexList& indexes) const;

  QModelIndex indexFromIndex (const QModelIndex& index) const
    { return indexForNode (nodeFromIndex (index)); }
  QModelIndex containerIndex (const QModelIndex& index) const
    { return nodeFromIndex (index) -> isContainer() ? index : parent (index); }

  /** Processes root index update. */
  virtual void newRootIndex (const QModelIndex& index, const QModelIndex& old);
  /** Processes expanded state update. */
  //virtual void expanded (const QModelIndex& index);

  virtual QModelIndex index (int row, int column, const QModelIndex& parent = QModelIndex()) const;
  virtual QModelIndex parent (const QModelIndex& child) const;

  virtual int rowCount (const QModelIndex& parent = QModelIndex()) const;
  virtual int columnCount (const QModelIndex& parent = QModelIndex()) const;

  virtual QVariant data (const QModelIndex& index, int role = Qt::DisplayRole) const;
  virtual bool setData (const QModelIndex& index, const QVariant& value, int role = Qt::EditRole);

  virtual QVariant headerData (int section, Qt::Orientation orientation, int role = Qt::DisplayRole) const;

  virtual bool hasChildren (const QModelIndex& index = QModelIndex()) const;
  virtual Qt::ItemFlags flags (const QModelIndex& index) const;

  virtual QStringList mimeTypes (void) const;
  virtual QMimeData* mimeData (const QModelIndexList& indexes) const;
  virtual bool dropMimeData (const QMimeData* data, Qt::DropAction action, int row, int column, const QModelIndex& parent);
  virtual Qt::DropActions supportedDropActions (void) const;

  void populateIndex (const QModelIndex& index);
  void vacateIndex (const QModelIndex& index);

  bool ancestor (const QModelIndex& index, const QModelIndex& container) const;

  void move (const QModelIndexList& list, const QModelIndex& after);
  void remove (const QModelIndexList& list);

protected slots:
  /** Adds the given nodes to the list. */
  void added (KPlayerContainerNode* parent, const KPlayerNodeList& nodes, KPlayerNode* after = 0);
  /** Removes the given nodes from the list. */
  void removed (KPlayerContainerNode* parent, const KPlayerNodeList& nodes);
  /** Updates the given node attributes. */
  void updated (KPlayerContainerNode* parent, KPlayerNode* node);

protected:
  void connectNode (KPlayerContainerNode* node);
  void disconnectNode (KPlayerContainerNode* node);

  void populate (KPlayerContainerNode* node);
  void vacate (KPlayerContainerNode* node);

  virtual bool populated (KPlayerContainerNode* node) const = 0;
  virtual void populateNode (KPlayerContainerNode* node) = 0;
  virtual void vacateNode (KPlayerContainerNode* node) = 0;

  virtual KPlayerNode* nodeAt (KPlayerContainerNode* parent, int index) const = 0;
  virtual int nodeIndex (KPlayerContainerNode* parent, KPlayerNode* node) const = 0;
  virtual int nodeCount (KPlayerContainerNode* parent) const = 0;
  virtual bool hasNodes (KPlayerContainerNode* parent) const = 0;
  virtual int attributeCount (void) const = 0;
  virtual QString nodeData (KPlayerNode* node, int column) const = 0;
  virtual void setNodeData (KPlayerNode* node, int column, const QString& data) const = 0;
  virtual QString attributeLabel (int column) const = 0;

  QList<KPlayerContainerNode*> m_populated;
  bool m_expandable;
};

/**List model.
  *@author kiriuja
  */
class KPlayerListModel : public KPlayerNodeModel
{
  Q_OBJECT

public:
  /** Sets up the node model. */
  KPlayerListModel (QObject* parent = 0);
  /** Cleans up the node model. */
  virtual ~KPlayerListModel();

  /** Initializes the model. */
  virtual void initialize (void);
  /** Frees up resources. */
  virtual void terminate (void);

  /** Returns the parent list view widget. */
  KPlayerListView* parent (void) const
    { return (KPlayerListView*) QObject::parent(); }

  /** Returns the attributes. */
  KPlayerNodeAttributes* attributes (void) const
    { return m_attributes; }

  int columns (void) const
    { return attributes() -> attributeCount(); }
  const QString& attributeName (int column) const
    { return attributes() -> attributeNames().at (column); }

  /** Processes root index update. */
  virtual void newRootIndex (const QModelIndex& index, const QModelIndex& old);
  /** Processes expanded state update. */
  //virtual void expanded (const QModelIndex& index);

  virtual QVariant headerData (int section, Qt::Orientation orientation, int role = Qt::DisplayRole) const;

  /** Sorts the model by the given column in the given order. */
  virtual void sort (int column, Qt::SortOrder order = Qt::AscendingOrder);

protected slots:
  /** Makes sure the view header is updated. */
  void updateColumns (void);
  /** Remembers the new column order. */
  void columnMoved (int, int, int);

signals:
  void columnsUpdated (void);

protected:
  virtual bool populated (KPlayerContainerNode* node) const;
  virtual void populateNode (KPlayerContainerNode* node);
  virtual void vacateNode (KPlayerContainerNode* node);

  virtual KPlayerNode* nodeAt (KPlayerContainerNode* parent, int index) const;
  virtual int nodeIndex (KPlayerContainerNode* parent, KPlayerNode* node) const;
  virtual int nodeCount (KPlayerContainerNode* parent) const;
  virtual bool hasNodes (KPlayerContainerNode* parent) const;
  virtual int attributeCount (void) const;
  virtual QString nodeData (KPlayerNode* node, int column) const;
  virtual void setNodeData (KPlayerNode* node, int column, const QString& data) const;
  virtual QString attributeLabel (int column) const;

  /** Attributes. */
  KPlayerNodeAttributes* m_attributes;
};

/**Tree model.
  *@author kiriuja
  */
class KPlayerTreeModel : public KPlayerNodeModel
{
  Q_OBJECT

public:
  /** Sets up the node model. */
  KPlayerTreeModel (QObject* parent = 0);
  /** Cleans up the node model. */
  virtual ~KPlayerTreeModel();

  /** Initializes the model. */
  virtual void initialize (void);
  /** Frees up resources. */
  virtual void terminate (void);

  /** Returns the parent tree view widget. */
  KPlayerTreeView* parent (void) const
    { return (KPlayerTreeView*) QObject::parent(); }

protected:
  virtual bool populated (KPlayerContainerNode* node) const;
  virtual void populateNode (KPlayerContainerNode* node);
  virtual void vacateNode (KPlayerContainerNode* node);

  virtual KPlayerNode* nodeAt (KPlayerContainerNode* parent, int index) const;
  virtual int nodeIndex (KPlayerContainerNode* parent, KPlayerNode* node) const;
  virtual int nodeCount (KPlayerContainerNode* parent) const;
  virtual bool hasNodes (KPlayerContainerNode* parent) const;
  virtual int attributeCount (void) const;
  virtual QString nodeData (KPlayerNode* node, int column) const;
  virtual void setNodeData (KPlayerNode* node, int column, const QString& data) const;
  virtual QString attributeLabel (int column) const;
};

/**Node controller.
  *@author kiriuja
  */
class KPlayerNodeController : public QObject
{
  Q_OBJECT

public:
  /** Sets up the node controller. */
  KPlayerNodeController (KActionCollection* ac, KPlayerPlaylist* playlist, KPlayerLibrary* parent);
  /** Cleans up the node controller. */
  virtual ~KPlayerNodeController();

  /** Returns the parent library widget. */
  KPlayerLibrary* parent (void) const
    { return (KPlayerLibrary*) QObject::parent(); }

  /** Returns the list view. */
  KPlayerListView* listView (void) const;
  /** Returns the tree view. */
  KPlayerTreeView* treeView (void) const;

  /** Configuration. */
  KPlayerConfiguration* configuration (void) const
    { return KPlayerEngine::engine() -> configuration(); }

  /** Returns the playlist. */
  KPlayerPlaylist* playlist (void) const
    { return m_playlist; }

  /** Returns the list of selected nodes. */
  KPlayerNodeList selection (void) const;

  /** Returns the active container node. */
  KPlayerContainerNode* activeContainer (void) const;

  /** Returns the action collection. */
  KActionCollection* actionCollection (void) const
    { return m_ac; }

  /** Retrieves an action from the action collection by name. */
  QAction* action (const char* name) const
    { return m_ac -> action (name); }

  /** Returns the playlist action list. */
  KPlayerContainerActionList* playlistActionList (void) const
    { return m_playlists; }

  /** Returns the column action list. */
  KPlayerToggleActionList* columnActionList (void) const
    { return m_columns; }

  /** Returns the edit action list. */
  KPlayerSimpleActionList* editActionList (void) const
    { return m_edit; }

  /** Returns the go to action list. */
  KPlayerContainerActionList* goToActionList (void) const
    { return m_go_to; }

  /** Returns the history action list. */
  KPlayerHistoryActionList* historyActionList (void) const
    { return m_history_list; }

  /** Initializes the controller. */
  void initialize (void);
  /** Frees up resources. */
  void terminate (void);

  /** Connects actions to their slots. */
  void connectActions (void);
  /** Disconnects and disables actions. */
  void disconnectActions (void);

  /** Adds the given list of URLs to a new playlist. */
  void addToNewPlaylistUrls (const KUrl::List& urls);
  /** Adds the given list of URLs to the root playlist. */
  void addToPlaylistsUrls (const KUrl::List& urls);
  /** Adds the given list of URLs to the collection. */
  void addToCollectionUrls (const KUrl::List& urls);

signals:
  /** Emitted when an action group is enabled or disabled. */
  void enableActionGroup (const QString& name, bool enable);

public slots:
  /** Plays the current selection. */
  void play (void);
  /** Plays the current selection after the currently played item finishes playing. */
  void playNext (void);
  /** Queues the current selection for playing. */
  void queue (void);
  /** Queues the current selection for playing after the currently played item finishes playing. */
  void queueNext (void);

  /** Displays the Open File dialog and adds the chosen files to the selected folder. */
  void addFiles (void);
  /** Displays the Open URL dialog and adds the entered URL to the selected folder. */
  void addUrl (void);
  /** Prompts for a new folder name and adds a new subfolder to the selected folder. */
  void addGroup (void);

  /** Saves the current selection as a new playlist. */
  void addToNewPlaylist (void);
  /** Adds the current selection to the root playlist. */
  void addToPlaylists (void);
  /** Adds the current selection to an existing playlist. */
  void addToPlaylist (KPlayerNode*);
  /** Adds the current selection to the collection. */
  void addToCollection (void);

  /** Starts rename mode for the current item. */
  void editName (void);
  /** Opens the Properties dialog for the current item. */
  void properties (void);

  /** Moves the current selection up in the playlist. */
  void moveUp (void);
  /** Moves the current selection down in the playlist. */
  void moveDown (void);
  /** Removes the current selection. */
  void remove (void);

  /** Selects all items in the current folder. */
  void selectAll (void);

  /** Starts edit mode for the chosen field of the current item. */
  void editField (int index);
  /** Hides or shows the column at the given index. */
  void toggleColumn (int index);

  /** Shows library and opens the Now Playing node in it. */
  void editPlaylist (void);

  /** Goes to the previous URL in the history. */
  void goBack (void);
  /** Goes to the next URL in the history. */
  void goForward (void);
  /** Goes to the given URL in the history. */
  void goToHistory (int index);

  /** Opens the parent node of the current root node. */
  void goUp (void);
  /** Opens the selected folder. */
  void goDown (void);
  /** Opens the origin node of the active node. */
  void goToOrigin (void);

  /** Updates actions according to the current selection. */
  void updateActions (void);

protected:
  /** Adds the given nodes to a new playlist. */
  KPlayerContainerNode* addToNewPlaylist (const KPlayerNodeList& list);

  /** Playlist. */
  KPlayerPlaylist* m_playlist;
  /** Action collection. */
  KActionCollection* m_ac;
  /** Playlist action list. */
  KPlayerContainerActionList* m_playlists;
  /** Column action list. */
  KPlayerToggleActionList* m_columns;
  /** Edit action list. */
  KPlayerSimpleActionList* m_edit;
  /** Go to action list. */
  KPlayerContainerActionList* m_go_to;
  /** History action list. */
  KPlayerHistoryActionList* m_history_list;
};

/**The node view.
  *@author kiriuja
  */
class KPlayerNodeView : public QTreeView
{
  Q_OBJECT

public: 
  /** Sets up the node view widget. */
  KPlayerNodeView (QWidget* parent = 0);
  /** Frees up resources. */
  virtual ~KPlayerNodeView();

  /** Initializes the view. */
  virtual void initialize (void);
  /** Frees up resources. */
  virtual void terminate (void);

  /** Returns the node model. */
  KPlayerNodeModel* model (void) const
    { return (KPlayerNodeModel*) QTreeView::model(); }

  /** Configuration. */
  KConfig* config (void) const
    { return KPlayerEngine::engine() -> config(); }

  /** Sets the root index and populates the node. */
  virtual void setRootIndex (const QModelIndex& index);

  /** Returns the index of the first child. */
  QModelIndex topIndex (void) const
    { return model() -> index (0, 0, rootIndex()); }

  /** Returns the active node. */
  QModelIndex activeIndex (void) const
  {
    QModelIndex active (selectionModel() -> currentIndex());
    return active.isValid() ? active : rootIndex();
  }
  /** Returns the active container. */
  QModelIndex activeContainer (void) const
  {
    QModelIndex active (activeIndex());
    return active.isValid() ? model() -> containerIndex (active) : rootIndex();
  }

  /** Returns the list of selected indexes. */
  QModelIndexList selection (void) const;
  /** Returns selected sibling indexes starting at the given index. */
  QModelIndexList selection (const QModelIndex& index) const;
  /** Returns contiguous selected or unselected indexes starting at the given index. */
  QModelIndexList contiguousSelection (const QModelIndex& index, bool selected) const;

public slots:
  void populateIndex (const QModelIndex& index);
  void vacateIndex (const QModelIndex& index);

signals:
  /** Emitted when the view gets focus. */
  void activated (KPlayerNodeView*);

protected:
  /** Returns the last visible descendant of the given index. */
  QModelIndex lastVisibleDescendant (const QModelIndex& parent) const;

  /** Updates actions and connections when the view gets the focus. */
  virtual void focusInEvent (QFocusEvent* event);
  /** Updates actions and connections when the view loses the focus. */
  virtual void focusOutEvent (QFocusEvent* event);
};

/**The list view.
  *@author kiriuja
  */
class KPlayerListView : public KPlayerNodeView
{
  Q_OBJECT

public: 
  /** Sets up the list view widget. */
  KPlayerListView (QWidget* parent = 0);
  /** Frees up resources. */
  virtual ~KPlayerListView();

  /** Initializes the view. */
  virtual void initialize (void);
  /** Frees up resources. */
  virtual void terminate (void);

  /** Returns the node model. */
  KPlayerListModel* model (void) const
    { return (KPlayerListModel*) QTreeView::model(); }

  /** Sets the root index and populates the node. */
  virtual void setRootIndex (const QModelIndex& index);

protected slots:
  /** Makes sure the view header is updated. */
  void updateColumns (void);
  /** Remembers the new column size. */
  void columnResized (int, int, int);
};

/**The tree view.
  *@author kiriuja
  */
class KPlayerTreeView : public KPlayerNodeView
{
  Q_OBJECT

public: 
  /** Sets up the tree view widget. */
  KPlayerTreeView (QWidget* parent = 0);
  /** Frees up resources. */
  virtual ~KPlayerTreeView();

  /** Initializes the view. */
  virtual void initialize (void);
  /** Frees up resources. */
  virtual void terminate (void);

  /** Returns the node model. */
  KPlayerTreeModel* model (void) const
    { return (KPlayerTreeModel*) QTreeView::model(); }

  void setActiveIndex (const QModelIndex& index);
};

/**The library splitter widget.
  *@author kiriuja
  */
class KPlayerLibrary : public QSplitter
{
  Q_OBJECT

public: 
  /** Sets up the view widget. Adds tree and list children. */
  KPlayerLibrary (KActionCollection* ac, KPlayerPlaylist* playlist, QWidget* parent = 0);
  /** Frees up resources. */
  virtual ~KPlayerLibrary();

  /** Initializes library. */
  void initialize (QMenu* menu);
  /** Frees up resources and saves state. */
  void terminate (void);

  /** Returns the controller. */
  KPlayerNodeController* controller (void) const
    { return m_controller; }

  /** Returns the tree view. */
  KPlayerTreeView* treeView (void) const
    { return m_tree; }
  /** Returns the list view. */
  KPlayerListView* listView (void) const
    { return m_list; }

  /** Returns the parent dock widget. */
  KPlayerLibraryWindow* parent (void) const
    { return (KPlayerLibraryWindow*) QSplitter::parent(); }

  /** Returns the history. */
  KPlayerHistory* history (void) const
    { return m_history; }

  /** Returns the popup menu. */
  QMenu* popupMenu (void) const
    { return m_popup; }

  /** Configuration. */
  KConfig* config (void) const
    { return KPlayerEngine::engine() -> config(); }

  /** Configuration. */
  KPlayerConfiguration* configuration (void) const
    { return KPlayerEngine::engine() -> configuration(); }

  /** Returns the active node view. */
  KPlayerNodeView* activeView (void) const
    { return m_active_view; }

  /** Remembers the widget height. */
  void rememberHeight (void)
    { m_height = height(); }

  /** Returns the size hint for the workspace. */
  virtual QSize sizeHint (void) const;

signals:
  /** Emitted when the widget is resized. */
  void resized (void);
  /** Emitted when the library needs to be shown. */
  void makeVisible (void);

public slots:
  /** Shows library and opens the given node. */
  void open (KPlayerNode* node);
  /** Sets the active node view. */
  void setActiveView (KPlayerNodeView* view);

protected slots:
  /** Disconnects actions when parent is hidden. */
  void visibilityChanged (bool);
  /** Handles tree view selection change. */
  void selectionChanged (const QItemSelection&, const QItemSelection&);
  /** Handles tree view current item change. */
  void currentIndexChanged (const QModelIndex& index, const QModelIndex&);
  /** Handles list view index activation. */
  void listIndexActivated (const QModelIndex& index);
  /** Shows the context menu. */
  void showContextMenu (const QPoint& position);
  /** Sets timer to reset the active item if needed. */
  void contextMenuHidden (void);
  /** Resets the active index. */
  void resetActiveIndex (void);

protected:
  /** Remembers the library height. Emits the resized signal. */
  virtual void resizeEvent (QResizeEvent* event);
  /** Sets the focus to the view that last had focus. */
  virtual void focusInEvent (QFocusEvent* event);
  /** Does nothing. */
  virtual void focusOutEvent (QFocusEvent* event);

  /** Controller. */
  KPlayerNodeController* m_controller;
  /** Tree view. */
  KPlayerTreeView* m_tree;
  /** List view. */
  KPlayerListView* m_list;
  /** Last active view. */
  KPlayerNodeView* m_active_view;
  /** Popup menu. */
  QMenu* m_popup;
  /** Indicates whether the popup menu is shown. */
  bool m_popup_menu_shown;
  /** History. */
  KPlayerHistory* m_history;
  /** Preferred height. */
  int m_height;
};

/**The library window.
  *@author kiriuja
  */
class KPlayerLibraryWindow : public QDockWidget
{
  Q_OBJECT

public: 
  /** Constructor. */
  KPlayerLibraryWindow (KActionCollection* ac, KPlayerPlaylist* playlist, QWidget* parent);

  /** Returns the library. */
  KPlayerLibrary* library (void)
    { return (KPlayerLibrary*) widget(); }

  /** Returns whether the widget is visible. */
  bool visible (void) const
    { return ! isHidden() && m_visibility; }
  /** Returns whether the widget is docked and visible. */
  bool docked (void) const
    { return ! isFloating() && ! isHidden(); }

signals:
  /** Emitted when the widget is moved. */
  void moved (bool docked);
  /** Emitted when the widget is resized. */
  void resized (void);

protected slots:
  /** Sets the visibility flag. */
  void setVisibility (bool visibility);

protected:
  /** Emits the moved signal. */
  virtual void moveEvent (QMoveEvent*);
  /** Emits the resized signal. */
  virtual void resizeEvent (QResizeEvent*);

  /** Sets the focus to the library. */
  virtual void focusInEvent (QFocusEvent* event);
  /** Does nothing. */
  virtual void focusOutEvent (QFocusEvent* event);

  /** Visibility indicator. */
  bool m_visibility;
};

inline KPlayerListView* KPlayerNodeController::listView (void) const
{
  return parent() -> listView();
}

inline KPlayerTreeView* KPlayerNodeController::treeView (void) const
{
  return parent() -> treeView();
}

inline KPlayerContainerNode* KPlayerNodeController::activeContainer (void) const
{
  KPlayerNodeView* view = parent() -> activeView();
  return (KPlayerContainerNode*) view -> model() -> nodeFromIndex (view -> activeContainer());
}

#endif
